@@ -54,11 +54,11 @@ static void luv_thread_release_vm(lua_State* L) {
 
 static const char* luv_getmtname(lua_State *L, int idx) {
   const char* name;
-  lua_getmetatable(L, idx);
-  lua_pushstring(L, "__name");
-  lua_rawget(L, -2);
-  name = lua_tostring(L, -1);
-  lua_pop(L, 2);
+  if (lua_getmetatable(L, idx)) {
+    lua_getfield(L, -1, "__name");
+    name = lua_tostring(L, -1);
+    lua_pop(L, 2);
+  }
   return name;
 }
 
@@ -228,11 +228,33 @@ static int luv_thread_arg_error(lua_State *L) {
     lua_typename(L, type), pos);
 }
 
-static int thread_dump (lua_State *L, const void *b, size_t size, void *ud) {
-  luaL_Buffer *B = (luaL_Buffer *)ud;
-  (void)L;
+// Based on code from lstrlib.c in Lua 5.4
+//
+// luaL_buffinit might push stuff onto the stack, so it is called after
+// lua_dump to ensure that the function to dump is still on the top of the stack.
+//
+// Note: The luaL_buffinit call is done within the thread_dump function instead of
+// before the lua_dump call in luv_thread_dumped because that plays nicer with
+// how the stack is used by luaL_Buffer (when calling luaL_buffinit and then
+// pushing the function onto the stack before calling lua_dump, the function
+// can end up getting replaced with a userdata during the course of lua_dump;
+// this doesn't necessarily cause problems but it seems safer to just match what
+// Lua does for its string.dump function).
+struct luv_thread_Writer {
+  int init;
+  luaL_Buffer B;
+};
 
-  luaL_addlstring(B, (const char *)b, size);
+static int thread_dump (lua_State *L, const void *b, size_t size, void *ud) {
+  struct luv_thread_Writer *state = (struct luv_thread_Writer *)ud;
+  if (!state->init) {
+    state->init = 1;
+    luaL_buffinit(L, &state->B);
+  }
+  // Starting with Lua 5.5.0, lua_dump calls this function with b == NULL to signal
+  // the end of the dump. We don't have to worry about that, though, because size will
+  // also be zero in that case and luaL_addlstring is a no-op when size is zero.
+  luaL_addlstring(&state->B, (const char *)b, size);
   return 0;
 }
 
@@ -241,18 +263,35 @@ static int luv_thread_dumped(lua_State* L, int idx) {
     lua_pushvalue(L, idx);
   } else {
     int ret, top;
-    luaL_Buffer B;
+    struct luv_thread_Writer state;
+    state.init = 0;
 
-    // In Lua >= 5.4.3, luaL_buffinit pushes a value onto the stack, so it needs to be called
-    // here to ensure that the function is at the top of the stack during the lua_dump call
-    luaL_buffinit(L, &B);
+    // The goal is to return from this function with the only change to the stack
+    // being the dumped thread pushed onto the top.
+    //
+    // Different Lua versions have different behaviors around how lua_dump and
+    // luaL_Buffer handle the stack, so the approach used here avoids the need to deal
+    // with those discrepencies.
+    //
+    // The idea is:
+    // - Push `nil` onto the stack as a placeholder for the resulting dumped string
+    // - Call lua_dump and let it and luaL_pushresult push/pop from the stack however
+    //   they wish (we know they will never touch our placeholder).
+    // - Move the result of luaL_pushresult to our placeholder's index and set it as
+    //   the top before returning
+    lua_pushnil(L);
+    top = lua_gettop(L);
+
+    // lua_dump needs the function at the top of the stack
     luaL_checktype(L, idx, LUA_TFUNCTION);
     lua_pushvalue(L, idx);
-    top = lua_gettop(L);
-    ret = lua_dump(L, thread_dump, &B, 1);
-    lua_remove(L, top);
+    ret = lua_dump(L, thread_dump, &state, 1);
     if (ret==0) {
-      luaL_pushresult(&B);
+      luaL_pushresult(&state.B);
+      // Move the result to our placeholder index and pop off whatever remains from
+      // the lua_dump call.
+      lua_replace(L, top);
+      lua_settop(L, top);
     } else
       luaL_error(L, "Error: unable to dump given function");
   }
